const MaxN     = ...;  (* MaxN = 2**p - 1 *)
      MaxNdiv2 = ...;  (* MaxNdiv2 = 2**(p-1) *)

type ComplexArray = array [0..MaxN] of complex;

procedure FFT(var x: ComplexArray; n: integer; PRofU: complex);
  (* Given a vector, x = [x(0) x(1) ...  x(n-1)], return the discrete
     Fourier transform of x, overwriting x.
     The procedure assumes that n is a power of 2.
     PRofU is a primitive nth root of unity. *)
  var RofU: array [0..MaxNdiv2] of complex;
        (* The extra element is to avoid subscript-out-of-range errors. *)
      omega, temp: complex;
      i, j, k, l, m: integer;
  begin
    (* For efficiency reasons, precompute the roots of unity. *)
    RofU[0] := (1.0,0.0);
    RofU[1] := PRofU;
    for i := 2 to n div 2 - 1 do RofU[i] := RofU[i-1]*RofU[1];

    (* Shuffle the data so that the recombination phase is orderly. *)
    (* i and j are the bit reversals of each other to lg n bits. *)
    j := n div 2; (* bit reversal of 1 *)
    for i := 1 to n-2 do
      begin
        (* Swap the pair, but swap them only once. *)
        if i < j then begin temp := x[i]; x[i] := x[j]; x[j] := temp end;
        (* Compute the value of j corresponding to the next i--with a bitwise
           addition of 1 from high bit (left) to low bit (right).
           Normal machine hardware addition is used to do the same
           operation low bit (right) to high bit (left) on i. *)
        k := n div 2;
        while j div k = 1 do
          begin
            j := j - k;
            k := k div 2
          end;
        j := j + k
      end;

    (* Remarks on variables:
         k is the size of the subarrays;
         l is the spacing within the nth roots of unity at which the kth roots
           of unity are found;
         j runs down each position of u, v, and w;
         m runs, in parallel to j, down the kth roots of unity; and
         i allows the computation to be done bottom-up, level by level. *)
    k := 1; l := n div 2;
    while k < n do
      begin
        m := 0;
        omega := RofU[0];
        for j := 0 to k-1 do (* Combine u(j) and v(j). *)
          begin
            i := j;
            while i < n do
              (* The bottom-up approach allows us to proceed across vectors. *)
              begin
                temp := x[i];
                x[i]   := temp + omega*x[i+k];
                x[i+k] := temp - omega*x[i+k];
                i := i + 2*k
              end;
            m := m + l;
            omega := RofU[m] (* undefined value on last fetch *)
          end;
        k := 2*k; l := l div 2
      end
  end; (* FFT *)
